﻿using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using Stride.Core.Mathematics;

namespace VL.ForceDirected
{

    public enum AttractionForce
    {
        Eades,
        FruchtermanReingold,
        Spring
    }

    public enum RepulsionForce
    {
        Eades,
        FruchtermanReingold,
        Attractor
    }

    public class FDGraph
    {
        private List<FDNode> FNodes;
        private List<FDEdge> FEdges;
        private Random FRand;


        public float Speed { get; set; }
        public string RootID { get; set; }

        public FDGraph()
        {
            FNodes = new List<FDNode>();
            FEdges = new List<FDEdge>();

            FRand = new Random();
        }

        public void SetForces(AttractionForce attraction, RepulsionForce repulsion)
        {
            Forces.Attraction = attraction;
            Forces.Repulsion = repulsion;
        }

        public void SetEades(Vector3 eades)
        {
            Forces.ConstantsEades = eades;
        }

        public void SetFruchtermanReingold(Vector2 fruchtermanReingold)
        {
            Forces.ConstantsFR = fruchtermanReingold;
        }

        public void SetSpringAttractor(Vector4 springAttractor)
        {
            Forces.ConstantsSA = springAttractor;
        }

        public void Clear()
        {
            FEdges.Clear();
            FNodes.Clear();
        }

        //add an edge by two ids
        public void AddEdge(string f, string t)
        {
            if (f == t) return;
            if (EdgeExists(f, t)) return;

            var fn = GetNode(f);
            var tn = GetNode(t);

            if (fn == null)
            {
                fn = new FDNode(f, IDToPos(f));
                FNodes.Add(fn);
            }

            if (tn == null)
            {
                tn = new FDNode(t, IDToPos(t));
                FNodes.Add(tn);
            }

            FEdges.Add(new FDEdge(fn, tn));

        }

        //get node by id
        public FDNode GetNode(string id)
        {
            var nodeCount = FNodes.Count;
            for (int i = 0; i < nodeCount; i++)
            {
                if (FNodes[i].ID == id) return FNodes[i];
            }
            return null;
        }

        public bool EdgeExists(string f, string t)
        {
            var edgeCount = FEdges.Count;
            for (int i = 0; i < edgeCount; i++)
            {
                var e = FEdges[i];
                if ((e.From.ID == f && e.To.ID == t) ||
                   (e.From.ID == t && e.To.ID == f)) return true;

            }
            return false;
        }

        public void RemoveNode(FDNode n)
        {
            foreach (var e in n.Parent)
            {
                e.From.Children.Remove(e);
                FEdges.Remove(e);
            }
            foreach (var e in n.Children)
            {
                e.To.Parent.Remove(e);
                FEdges.Remove(e);
            }
            FNodes.Remove(n);
        }

        public List<FDNode> Nodes
        {
            get { return FNodes; }
        }

        public List<FDEdge> Edges
        {
            get { return FEdges; }
        }

        public void Update(float deltaTime)
        {
            var edgeCount = FEdges.Count;
            var nodeCount = FNodes.Count;

            //attraction along edges
            for (int i = 0; i < edgeCount; i++)
            {
                FEdges[i].Update();
            }

            //repulsion of the nodes
            for (int i = 0; i < nodeCount; i++)
            {
                var n = FNodes[i];
                n.Velocity += CalcRepulsion(n);
                n.Velocity *= Speed;
            }

            //update position
            for (int i = 0; i < nodeCount; i++)
            {
                var n = FNodes[i];

                n.Update(deltaTime);
            }
        }


        public void UpdateParallel(float deltaTime)
        {
            var edgeCount = FEdges.Count;
            var nodeCount = FNodes.Count;

            //attraction along edges
            Parallel.For(0, edgeCount, i =>
            {
                FEdges[i].Update();
            });

            //repulsion of the nodes
            Parallel.For(0, nodeCount, i =>
            {
                var n = FNodes[i];
                n.Velocity += CalcRepulsion(n);
                n.Velocity *= Speed;

                //update position
                n.Update(deltaTime);
            });

            //update position
            //Parallel.For(0, nodeCount, i =>
            //{

            //    var n = FNodes[i];

            //    n.Update(deltaTime);

            //});

        }

        private Vector3 CalcRepulsion(FDNode n)
        {
            var rep = new Vector3(0);

            if (n.Parent.Count == 0) return rep;

            var oCount = FNodes.Count;
            for (int i = 0; i < oCount; i++)
            {
                var o = FNodes[i];

                if (o != n)
                {
                    var diff = n.Position - o.Position;
                    rep += Forces.Repulse(diff);
                }
            }

            return rep / oCount;
        }



        protected Vector3 RandomVector(Vector3 pos) 
        {
            pos.X = (float)FRand.NextDouble();
            pos.Y = (float)FRand.NextDouble();
            pos.Z = (float)FRand.NextDouble();

            return pos;
        }

        protected Vector3 IDToPos(string id)
        {
            var pos = new Vector3(0f);

            if (id == RootID) return pos;

            return RandomVector(pos);


           // var hash = id.GetHashCode();

           // unchecked
           // {
           //     pos.X = ((hash & 0x00FF0000) >> 16) / 255;
           //     pos.Y = ((hash & 0x0000FF00) >> 8) / 255;
           //     pos.Z = ((hash & 0x000000FF)) / 255;
           // }

           //return pos * 4 - 2;
        }
    }

    //a node
    public class FDNode
    {
        public Vector3 Position;
        public Vector3 Velocity;
        public string ID;
        public List<FDEdge> Parent;
        public List<FDEdge> Children;

        public FDNode(string id, Vector3 pos)
        {
            ID = id;

            Position = pos;
            Velocity = new Vector3(0);

            Children = new List<FDEdge>();
            Parent = new List<FDEdge>();
        }

        public void Update(float deltaTime)
        { 
            //calc position
            Position += Velocity * deltaTime;
        }

    }

    //a graph edge
    public class FDEdge
    {
        public FDNode From;
        public FDNode To;

        public FDEdge(FDNode f, FDNode t)
        {
            From = f;
            To = t;
            From.Children.Add(this);
            To.Parent.Add(this);
        }

        public void Update()
        {
            var att = Forces.Attract(From.Position - To.Position);

            From.Velocity -= From.Parent.Count == 0 ? new Vector3(0f) : att;
            To.Velocity += att;
        }
    }

    //the force functions
    public static class Forces
    {
        //constants
        public static AttractionForce Attraction;
        public static RepulsionForce Repulsion;

        public static Vector3 ConstantsEades;
        public static Vector2 ConstantsFR;
        public static Vector4 ConstantsSA;

        //public force functions
        public static Vector3 Attract(Vector3 diff)
        {
            switch (Attraction)
            {
                case AttractionForce.Eades: return AttractEades(diff);
                case AttractionForce.FruchtermanReingold: return AttractFR(diff);
                default: return AttractSpring(diff);
            }

        }

        public static Vector3 Repulse(Vector3 diff)
        {
            switch (Repulsion)
            {
                case RepulsionForce.Eades: return RepulseEades(diff);
                case RepulsionForce.FruchtermanReingold: return RepulseFR(diff);
                default: return RepulseAttractor(diff);
            }
        }

        //Eades
        private static Vector3 AttractEades(Vector3 diff)
        {
            var l = Math.Abs(Vector3.Dot(diff,diff));
            diff /= l;

            return diff * ConstantsEades.X * (float) Math.Log(l / ConstantsEades.Y);
        }

        private static Vector3 RepulseEades(Vector3 diff)
        {
            var l = Math.Abs(Vector3.Dot(diff, diff));
            diff /= l;
            return diff * Math.Min(ConstantsEades.Z / l, 10f);
        }

        //Fruchterman-Reingold
        private static Vector3 AttractFR(Vector3 diff)
        {
            var l = diff.Length();
            diff /= l;

            return diff * ((l * l) / ConstantsFR.X);
        }

        private static Vector3 RepulseFR(Vector3 diff)
        {
            var l = diff.Length();
            diff /= l;
            return diff * ((ConstantsFR.Y * ConstantsFR.Y) / l);
        }

        //SpringAttractor
        private static Vector3 AttractSpring(Vector3 diff)
        {
            var l = diff.Length();
            diff /= l;

            return diff * (l - ConstantsSA.X) * 0.5f;
        }

        private static Vector3 RepulseAttractor(Vector3 diff)
        {
            var l = diff.Length();

            if (l > ConstantsSA.Y) return new Vector3(0f);

            diff /= l;
            var s = (float)Math.Pow(l / ConstantsSA.Y, 1 / ConstantsSA.Z);
            var f = s * 9f * ConstantsSA.W * (1 / (s + 1) + ((s - 3) / 4)) / l;
            return diff * f;
        }
    }


}
